@@ -1,7 +1,11 @@
 import ast
+import base64
 import inspect
 import re
 import os
+import logging
+import random
+import string
 import sys
 
 from django import forms
@@ -15,15 +19,40 @@ from django.template.defaultfilters import filesizeformat
 from django.utils.module_loading import import_string
 from django.utils.translation import gettext as _
 
+from . enc import encrypt
+from . formsets import build_formset
+from . widgets import *
+from . settings import *
+from . utils import (_split_choices,
+                     _split_choices_in_list_canc,
+                     _successivo_ad_oggi)
+
+
 try:
     from filesig.filesig import get_signatures
 except:
     from filesig import get_signatures
 
-from . settings import *
-from . utils import (_split_choices,
-                     _split_choices_in_list_canc,
-                     _successivo_ad_oggi)
+
+logger = logging.getLogger(__name__)
+
+
+
+MAX_UPLOAD_SIZE = getattr(settings, 'MAX_UPLOAD_SIZE', MAX_UPLOAD_SIZE)
+ATTACH_NAME_MAX_LEN = getattr(settings, 'ATTACH_NAME_MAX_LEN',
+                              ATTACH_NAME_MAX_LEN)
+PERMITTED_UPLOAD_FILETYPE = getattr(settings,
+                                    'PERMITTED_UPLOAD_FILETYPE',
+                                    PERMITTED_UPLOAD_FILETYPE)
+WRONG_TYPE = getattr(settings, 'WRONG_TYPE', WRONG_TYPE)
+WRONG_SIZE = getattr(settings, 'WRONG_SIZE', WRONG_SIZE)
+WRONG_LENGHT = getattr(settings, 'WRONG_LENGHT', WRONG_LENGHT)
+IMG_FILETYPE = getattr(settings, 'IMG_FILETYPE', IMG_FILETYPE)
+DATA_FILETYPE = getattr(settings, 'DATA_FILETYPE', DATA_FILETYPE)
+PDF_FILETYPE = getattr(settings, 'PDF_FILETYPE', PDF_FILETYPE)
+P7M_FILETYPE = getattr(settings, 'P7M_FILETYPE', P7M_FILETYPE)
+CLASSIFICATION_LIST = getattr(settings, 'CLASSIFICATION_LIST', CLASSIFICATION_LIST)
+
 
 def get_fields_types(class_name=sys.modules[__name__]):
     fields_types = []
@@ -35,6 +64,7 @@ def get_fields_types(class_name=sys.modules[__name__]):
     fields_types.sort(key=lambda tup: tup[1])
     return fields_types
 
+
 def format_field_name(field_name, lower=True):
     f = field_name.replace(' ','_')
     f = f.replace('(','')
@@ -43,48 +73,6 @@ def format_field_name(field_name, lower=True):
     if lower: return f.lower()
     return f.upper
 
-def get_empty_form(form_class=forms.Form):
-    class Dynamic(form_class):
-        pass
-    return Dynamic
-
-def build_formset(choices, extra=0, required=False, prefix='form', data=None):
-    """ Get formset
-    """
-    _regexp = '(?P<colname>[a-zA-Z0-9_]*)\((?P<coldict>[\{\}\.0-9a-zA-Z\'\"\:\;\_\,\s\-]*)\)'
-    min_num = 0
-    if required: min_num = 1
-    eform = get_empty_form()
-    for choice in choices:
-        colname = choice # needed for simple CharField withoud attrs
-        contenuto = re.search(_regexp, choice)
-        field_dict = None
-        if contenuto:
-            coldict = contenuto.groupdict().get('coldict')
-            colname = contenuto.groupdict()['colname']
-            if coldict:
-                field_dict = ast.literal_eval(coldict)
-                field_type_name = field_dict['type']
-                del field_dict['type']
-                custom_field = getattr(sys.modules[__name__], field_type_name)(**field_dict) \
-                               if hasattr(sys.modules[__name__], field_type_name) else CustomCharField()
-                custom_widget = getattr(settings, 'CUSTOM_WIDGETS').get(field_type_name) \
-                                if hasattr(settings, 'CUSTOM_WIDGETS') else None
-                if custom_widget:
-                    custom_field.widget = import_string(custom_widget)()
-
-                if field_dict.get('choices'):
-                    custom_field.choices += _split_choices(field_dict.get('choices'))
-        else:
-            custom_field = CustomCharField()
-        eform.base_fields[colname] = custom_field
-        eform.declared_fields[colname] = custom_field
-
-    # Django formset
-    fac =  formset_factory(eform, extra=extra, min_num=min_num)
-    if data: return fac(prefix=prefix, data=data)
-    return fac(prefix=prefix)
-
 
 class BaseCustomField(Field):
     """
@@ -164,10 +152,10 @@ class CustomFileField(FileField, BaseCustomField):
         if data:
             msg = ''
             self_valid_extensions = getattr(self, 'valid_extensions') if hasattr(self, 'valid_extensions') else None
-            settings_upload_filetype = getattr(settings, 'PERMITTED_UPLOAD_FILETYPE') if hasattr(settings, 'PERMITTED_UPLOAD_FILETYPE') else PERMITTED_UPLOAD_FILETYPE
+            settings_upload_filetype = PERMITTED_UPLOAD_FILETYPE
             permitted_upload_filetype = self_valid_extensions or settings_upload_filetype
-            max_upload_size = getattr(settings, 'MAX_UPLOAD_SIZE') if hasattr(settings, 'MAX_UPLOAD_SIZE') else MAX_UPLOAD_SIZE
-            attach_max_len = getattr(settings, 'ATTACH_NAME_MAX_LEN') if hasattr(settings, 'ATTACH_NAME_MAX_LEN') else ATTACH_NAME_MAX_LEN
+            max_upload_size = MAX_UPLOAD_SIZE
+            attach_max_len = ATTACH_NAME_MAX_LEN
 
             if data.content_type not in permitted_upload_filetype:
                 msg_tmpl = WRONG_TYPE
@@ -190,7 +178,7 @@ class CustomImageField(CustomFileField):
     field_type = _("Allegato Immagine")
 
     def __init__(self, *args, **data_kwargs):
-        self.valid_extensions = getattr(settings, 'IMG_FILETYPE') if hasattr(settings, 'IMG_FILETYPE') else IMG_FILETYPE
+        self.valid_extensions = IMG_FILETYPE
         super().__init__(*args, **data_kwargs)
 
 
@@ -201,7 +189,7 @@ class CustomDataField(CustomFileField):
     field_type = _("Allegato file dati (JSON, CSV, Excel)")
 
     def __init__(self, *args, **data_kwargs):
-        self.valid_extensions = getattr(settings, 'DATA_FILETYPE') if hasattr(settings, 'DATA_FILETYPE') else DATA_FILETYPE
+        self.valid_extensions = DATA_FILETYPE
         super().__init__(*args, **data_kwargs)
 
 
@@ -212,7 +200,7 @@ class CustomPDFField(CustomFileField):
     field_type = _("Allegato PDF")
 
     def __init__(self, *args, **data_kwargs):
-        self.valid_extensions = getattr(settings, 'PDF_FILETYPE') if hasattr(settings, 'PDF_FILETYPE') else PDF_FILETYPE
+        self.valid_extensions = PDF_FILETYPE
         super().__init__(*args, **data_kwargs)
 
 
@@ -255,7 +243,7 @@ class CustomSignedPdfField(CustomSignedFileField):
     fileformat = 'pdf'
 
     def __init__(self, *args, **data_kwargs):
-        self.valid_extensions = getattr(settings, 'PDF_FILETYPE') if hasattr(settings, 'PDF_FILETYPE') else PDF_FILETYPE
+        self.valid_extensions = PDF_FILETYPE
         super().__init__(*args, **data_kwargs)
 
 
@@ -266,7 +254,7 @@ class CustomSignedP7MField(CustomSignedFileField):
     fileformat = 'p7m'
 
     def __init__(self, *args, **data_kwargs):
-        self.valid_extensions = getattr(settings, 'P7M_FILETYPE') if hasattr(settings, 'P7M_FILETYPE') else P7M_FILETYPE
+        self.valid_extensions = P7M_FILETYPE
         super().__init__(*args, **data_kwargs)
 
 
@@ -317,7 +305,7 @@ class TextAreaField(CharField, BaseCustomField):
     TextArea
     """
     field_type = _("Testo lungo")
-    widget = forms.Textarea()
+    widget = forms.Textarea
 
 
 class CheckBoxField(BooleanField, BaseCustomField):
@@ -325,7 +313,7 @@ class CheckBoxField(BooleanField, BaseCustomField):
     BooleanField Checkbox
     """
     field_type = _("Checkbox")
-    widget = forms.CheckboxInput()
+    widget = forms.CheckboxInput
 
 
 class CustomSelectBoxField(CustomChoiceField):
@@ -344,14 +332,14 @@ class CustomRadioBoxField(CustomChoiceField):
     CheckBox multiplo
     """
     field_type = _("Lista di opzioni (checkbox)")
-    widget = forms.RadioSelect()
+    widget = forms.RadioSelect
 
 class BaseDateField(DateField, BaseCustomField):
     """
     DateField
     """
     field_type = _("Data")
-    widget = forms.DateInput()
+    widget = forms.DateInput
     input_formats = settings.DATE_INPUT_FORMATS
 
 
@@ -513,11 +501,98 @@ class CustomHiddenField(CharField, BaseCustomField):
     CharField Hidden
     """
     field_type = _("Campo nascosto")
+    widget = forms.HiddenInput
 
     def define_value(self, custom_value, **kwargs):
         self.widget = forms.HiddenInput(attrs={'value': custom_value})
 
 
+class CaptchaHiddenField(CustomHiddenField):
+    """
+    Captcha Hidden field
+    """
+    field_type = _("Campo nascosto captcha")
+
+
+class CaptchaField(BaseCustomField):
+    """
+    Captcha
+
+    in settings:
+        CAPTCHA_FONTS = ['/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf',
+                         '/usr/share/fonts/truetype/liberation/LiberationMono-Italic.ttf']
+        CAPTCHA_SECRET = '6sa78d6as83$_RDF'
+        CAPTCHA_SALT = 'ingoalla'
+
+    """
+    field_type = _("_Captcha")
+    widget = CaptchaWidget
+
+    def define_value(self, custom_value, **kwargs):
+        self.widget = CaptchaWidget(attrs={'value': custom_value,
+                                           'hidden_field': kwargs['hidden_field']})
+
+
+class CustomCaptchaComplexField(BaseCustomField):
+    """
+    Captcha
+
+    in settings:
+        CAPTCHA_FONTS = ['/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf',
+                         '/usr/share/fonts/truetype/liberation/LiberationMono-Italic.ttf']
+        CAPTCHA_SECRET = '6sa78d6as83$_RDF'
+        CAPTCHA_SALT = 'ingoalla'
+
+    """
+    field_type = _("Captcha")
+    is_complex = True
+
+    def __init__(self, *args, **kwargs):
+        # CaPTCHA
+        parent_label = kwargs.get('label')
+
+        self.captcha_hidden = CaptchaHiddenField()
+        length = getattr(settings, 'CAPTCHA_LENGTH', 5)
+        text = ''.join([random.choice(string.ascii_letters) for i in range(length)])
+        value = base64.b64encode(encrypt(text)).decode()
+        self.captcha_hidden.define_value(custom_value=value)
+        logger.debug(text, value)
+
+        self.captcha_hidden.required = True
+        self.captcha_hidden.name = "{}_hidden_dyn".format(format_field_name(parent_label))
+        self.captcha_hidden.parent = self
+        self.captcha_hidden.label = ''
+
+        self.captcha = CaptchaField(*args, **kwargs)
+        self.captcha.required = True
+        self.captcha.define_value(custom_value=text,
+                                  hidden_field="id_{}".format(self.captcha_hidden.name))
+        self.captcha.label = parent_label
+        self.captcha.name = "{}_dyn".format(format_field_name(parent_label))
+        self.captcha.help_text = _("CaPTCHA: insert the value represented in the Image")
+        self.captcha.parent = self
+        # super().__init__(*args, **kwargs)
+
+    def get_fields(self):
+        return [self.captcha, self.captcha_hidden]
+
+    def raise_error(self, name, cleaned_data, **kwargs):
+        """
+        """
+        _err_msg = _('CaPTCHA not valid!')
+        errors = []
+        value = cleaned_data.get(self.captcha.name)
+        check = cleaned_data.get(self.captcha_hidden.name)
+        try:
+            cvalue = decrypt(check).decode()
+        except:
+            errors.append(_err_msg)
+
+        if value and cvalue.lower() != value.lower():
+            errors.append(_err_msg)
+        return errors
+
+
 class CustomComplexTableField(ChoiceField, BaseCustomField):
     """
     CustomComplexTableField
